import logging
import requests
from enum import IntEnum

from odoo.addons.saas_worker.const import GANDI_API_KEY, GANDI_ODOO_EMPLOYEE_ID, GANDI_ORG_ID, TIMEOUT
from odoo.exceptions import ValidationError


_logger = logging.getLogger(__name__)


class GandiAuth(requests.auth.AuthBase):
    def __init__(self, api_key):
        self.api_key = api_key

    def __call__(self, r):
        r.headers['authorization'] = 'Apikey %s' % self.api_key
        return r


# https://doc.rpc.gandi.net/domain/reference.html#ContactType
class ContactType(IntEnum):
    person = 0
    company = 1
    association = 2
    public_body = 3
    reseller = 4

    def __str__(self):
        return self.name

class GandiContact(dict):
    def __init__(self, data):
        # Use __init__ directly if parsing data retrieved from Gandi.
        self.type = data['type']
        self.given = data['given']
        self.family = data['family']
        self.email = data['email']
        self.phone = data['phone']
        self.country = data['country']
        self.streetaddr = data['streetaddr']
        self.city = data['city']
        self.zip = data['zip'] if self.country != 'BO' else '0000'
        self.data_obfuscated = True
        self.mail_obfuscated = True
        self.extra_parameters = data.get('extra_parameters', {})

        if data.get('orgname'):
            self.orgname = data['orgname']
        if data.get('state'):
            self.state = data['state']

    @staticmethod
    def odoo_data():
        return {
            'given': 'Fabien',
            'family': 'Pinckaers',
            'email': 'gandi@odoo.com',
            'phone': '+32 (0)81 81 37 00',
            'country': 'BE',
            'streetaddr': 'Chaussée de Namur, 40',
            'city': 'Grand-Rosière',
            'zip': '1367',
        }

    @classmethod
    def odoo_employee(cls):
        return cls(dict(cls.odoo_data(), **{
            'type': ContactType.person.value,
            'extra_parameters': {
                # https://doc.rpc.gandi.net/domain/reference.html#ESTipoIdent
                'x-es_tipo_identification': '0',
                'x-es_identification': GANDI_ODOO_EMPLOYEE_ID,
            },
        }))

    @classmethod
    def odoo_sa(cls):
        return cls(dict(cls.odoo_data(), **{
            'type': ContactType.company.value,
            'orgname': 'Odoo SA',
        }))

    @classmethod
    def owner(cls, fqdn, data):
        # The owner must pass all these validations before registration.
        fields = ['type', 'given', 'family', 'email', 'phone', 'country', 'streetaddr', 'city']
        if data['country'] != 'BO':
            fields.append('zip')
        for f in fields:
            if data.get(f, '') == '':
                raise ValueError(f'Field {f} is required.')

        if data['type'] not in list(ContactType):
            raise ValueError('Unknown contact type: %s.' % (data['type']))

        if data['type'] != ContactType.person.value and not data.get('orgname'):
            raise ValueError('Field orgname is required.')

        # ccTLD specific fields.
        is_org = data['type'] != ContactType.person.value
        field_set = set()

        if fqdn.endswith('.ca'):
            field_set.add('ca_legaltype')
            if is_org:
                field_set.add('ca_business_entity_type')
            if data['country'] == 'CA' and data.get('state', '') == '':
                raise ValueError('Field state is required.')
        elif fqdn.endswith('.es'):
            field_set.update(['es_tipo_identification', 'es_identification', 'es_legalform'])
            if data['country'] == 'ES' and data.get('state', '') == '':
                raise ValueError('Field state is required.')
        elif fqdn.endswith('.fi'):
            field_set.add('fi_isfinish')
            if is_org:
                field_set.update(['fi_business_number', 'fi_organization_type'])
            else:
                if data.get('fi_isfinish', False):
                    field_set.add('fi_ident_number')
                else:
                    field_set.add('fi_birth_date')
        elif fqdn.endswith('.it'):
            if data['country'] == 'IT' and data.get('state', '') == '':
                raise ValueError('Field state is required.')
            if is_org:
                field_set.update(['it_vat_number', 'it_registrant_entity_type'])
            field_set.add('it_pin')
        elif fqdn.endswith('.se'):
            field_set.add('se_ident_number')
            if is_org:
                field_set.add('se_registrant_vatid')
        elif fqdn.endswith(('.co.uk', '.uk')):
            field_set.add('uk_contact_type')
            if data.get('uk_contact_type', '') in ['LTD', 'PLC', 'LLP', 'IP', 'SCH', 'RCHAR']:
                field_set.add('uk_co_no')
        elif fqdn.endswith('.us'):
            field_set.update(['us_app_purpose', 'us_nexus_category'])

        bool2str_fields = ['fi_isfinish']  # bool need to become yes/no str
        # These fields should not be prefixed with 'x-{cctld}_'.
        alter_name = {
            'fi_birth_date': 'birth_date',
            'it_vat_number': 'vat_number',
        }

        contact = cls(data)

        for f in field_set:
            # False and 0 are valid values, empty string not.
            if data.get(f, '') == '':
                raise ValueError(f'Field {f} is required.')
            if f in bool2str_fields:
                data[f] = data[f] and 'yes' or 'no'
            key = f in alter_name and alter_name[f] or f'x-{f}'
            # e.g. self.extra_parameters['x-ca_business_entity_type'] = data['ca_business_entity_type']
            contact.extra_parameters[key] = data[f]

        return contact


class GandiError(ValidationError):
    def __init__(self, errors):
        super().__init__("Bad request")
        self.errors = errors


class GandiAPI:
    def __init__(self, api_key, org_id):
        self.endpoint = 'https://api.gandi.net/v5'
        self._auth = GandiAuth(api_key)
        self.org_id = org_id

    def status(self, domain, process='create'):
        # https://api.gandi.net/docs/domains/#v5-domain-check

        url = '%s/domain/check' % self.endpoint
        querystring = {
            'currency': 'EUR',
            'name': domain.lower(),
            'sharing_id': self.org_id,
            'processes': process,
        }

        response = self._request('GET', url, params=querystring)

        return response.json()

    def register_domain(self, fqdn, owner_data, dry_run=False):
        # https://api.gandi.net/docs/domains/#post-v5-domain-domains

        url = "%s/domain/domains" % self.endpoint
        querystring = {'sharing_id': self.org_id}

        # Odoo SA is the administrative, billing and technical contact unless a physical person is required.
        admin = GandiContact.odoo_sa()
        bill = GandiContact.odoo_sa()
        tech = GandiContact.odoo_sa()

        # Some ccTLDs require the contacts to be a physical person.
        if fqdn.endswith(('.ca', '.es')):
            admin = GandiContact.odoo_employee()
            tech = GandiContact.odoo_employee()
        if fqdn.endswith('.es'):
            bill = GandiContact.odoo_employee()

        owner = GandiContact.owner(fqdn, owner_data)

        duration = 2 if fqdn.endswith('.gr') else 1
        payload = {
            'fqdn': fqdn,
            'duration': duration,
            'owner': vars(owner),
            'admin': vars(admin),
            'bill': vars(bill),
            'tech': vars(tech),
            'currency': 'EUR',
        }

        headers = {'dry-run': '1'} if dry_run else None

        try:
            response = self._request('POST', url, headers=headers, json=payload, params=querystring)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 400:
                resp = e.response.json()
                errors = resp.get('errors', [])
                # Some endpoints use 'message' for reporting errors despite the documentation suggesting otherwise.
                if not errors and resp.get('message', ''):
                    errors.append({'description': resp['message']})
                raise GandiError(errors) from e
            raise

        data = response.json()

        # HTTP 200 OK with errors in response body.
        err = data.get('errors', [])
        if err:
            raise GandiError(err)

        # Dry-run response differs from production response.
        if dry_run and data['status'] != 'success':
            _logger.error('gandi: failed to register: domain=%s errors=%s', fqdn, data.get('errors', []))
            raise ValidationError(' '.join([err['description'] for err in data.get('errors', [])]))

        # Production response.
        if not dry_run and response.status_code == 202:
            _logger.info('gandi: domain name registered: fqdn=%s, message=%s', fqdn, data['message'])

        return data

    def info(self, domain):
        # https://api.gandi.net/docs/domains/#get-v5-domain-domains-domain

        url = '%s/domain/domains/%s' % (self.endpoint, domain)

        response = self._request('GET', url)

        return response.json()

    def get_contacts(self, domain):
        # https://api.gandi.net/docs/domains/#get-v5-domain-domains-domain-contacts

        url = '%s/domain/domains/%s/contacts' % (self.endpoint, domain)

        response = self._request('GET', url)
        data = response.json()
        contacts = {
            'admin': GandiContact(data['admin']),
            'bill': GandiContact(data['bill']),
            'owner': GandiContact(data['owner']),
            'tech': GandiContact(data['tech']),
        }

        return contacts

    def tag_domain(self, domain, tags):
        # https://api.gandi.net/docs/domains/#put-v5-domain-domains-domain-tags

        url = "%s/domain/domains/%s/tags" % (self.endpoint, domain.lower())
        payload = {
            'tags': tags,
        }

        self._request('PUT', url, json=payload)

    def renew_domain(self, domain, dry_run=False):
        # https://api.gandi.net/docs/domains/#post-v5-domain-domains-domain-renew

        url = "%s/domain/domains/%s/renew" % (self.endpoint, domain.lower())
        querystring = {'sharing_id': self.org_id}
        headers = {'dry-run': '1'} if dry_run else None

        duration = 2 if domain.lower().endswith('.gr') else 1
        payload = {
            'duration': duration,
        }

        response = self._request('POST', url, headers=headers, json=payload, params=querystring)
        data = response.json()

        # Dry-run response differs from real response.
        if dry_run and data['status'] != 'success':
            _logger.error('gandi: failed to renew: errors=%s', data.get('errors', []))
            return False

        # Real response.
        if not dry_run and response.status_code == 202:
            _logger.info('gandi: domain name renewed: %s', data['message'])

        return True

    def get_records(self, domain):
        # https://api.gandi.net/docs/livedns/#get-v5-livedns-domains-fqdn-records

        url = '%s/livedns/domains/%s/records' % (self.endpoint, domain)
        return self._request('GET', url).json()

    def get_record_content(self, domain, rtype, name):
        # https://api.gandi.net/docs/livedns/#get-v5-livedns-domains-fqdn-records-rrset_name-rrset_type

        try:
            url = '%s/livedns/domains/%s/records/%s/%s' % (self.endpoint, domain, name, rtype)
            response = self._request('GET', url)
            data = response.json()
            return data['rrset_values']
        except requests.exceptions.HTTPError as e:
            # 404 means there is no DNS records for the name/type pair.
            if e.response.status_code == 404:
                return []
            raise

    def put_record(self, domain, rtype, name, content, ttl=None):
        # https://api.gandi.net/docs/livedns/#put-v5-livedns-domains-fqdn-records-rrset_name-rrset_type

        url = '%s/livedns/domains/%s/records/%s/%s' % (self.endpoint, domain, name, rtype)

        # An empty PUT does not work for MX records -> use DELETE when content is empty.
        method = 'DELETE'
        payload = None

        if content:
            method = 'PUT'
            payload = {
                'rrset_values': content,
            }

            if ttl:
                payload['rrset_ttl'] = ttl

        self._request(method, url, json=payload)

    def delete_records(self, domain):
        # https://api.gandi.net/docs/livedns/#delete-v5-livedns-domains-fqdn-records

        url = '%s/livedns/domains/%s/records' % (self.endpoint, domain)

        try:
            response = self._request('DELETE', url)
            return response.status_code == 204
        except Exception:
            return False

    def create_web_redirection(self, domain, host, redirect_to, redirect_type):
        # https://api.gandi.net/docs/domains/#post-v5-domain-domains-domain-webredirs

        url = '%s/domain/domains/%s/webredirs' % (self.endpoint, domain)

        payload = {
            'host': host,
            'override': True,
            'protocol': 'https',
            'type': redirect_type,
            'url': redirect_to,
        }

        try:
            response = self._request('POST', url, json=payload)
        except requests.exceptions.HTTPError as e:
            # Return a more user-friendly error.
            if e.response.status_code == 400 and e.response.json().get('cause', '') == 'CAUSE_EXIST':
                raise Exception('Redirection from %s already exists.' % (host)) from e
            raise
        return response.json()

    def resend_reachability_email(self, domain):
        # https://api.gandi.net/docs/domains/#v5-domain-domains-domain-reachability
        url = "%s/domain/domains/%s/reachability" % (self.endpoint, domain.lower())
        querystring = {'sharing_id': self.org_id}
        payload = {
            'action': 'resend',
        }

        self._request('PATCH', url, json=payload, params=querystring)

    def _request(self, method, url, headers=None, json=None, params=None):
        try:
            response = requests.request(method, url, auth=self._auth, headers=headers, json=json, params=params, timeout=TIMEOUT)
            _logger.info('gandi: %s [%s] responsed with %s - %s', response.url, method, response.text, response.status_code)
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            _logger.exception('gandi: http error: status=%s', response.status_code)
            raise
        except Exception as e:
            _logger.exception('gandi: bad request body: body=%s', json)
            raise Exception('We were unable to connect to the service.') from e
        else:
            return response


CCTLDS = [
    '.at', '.be', '.ca', '.co', '.ch', '.cz', '.de', '.es', '.fi', '.fr',
    '.gr', '.in', '.it', '.lu', '.nl', '.nz', '.pl', '.se', '.sk', '.uk',
    '.us', '.co.uk',
]
GTLDS = [
    '.app', '.biz', '.cloud', '.club', '.com', '.me', '.net', '.org', '.page',
    '.top', '.xyz',
]
TLDS = GTLDS + CCTLDS

GANDI = GandiAPI(GANDI_API_KEY, GANDI_ORG_ID)
